package org.colinvella.fancyfish.block;

import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.Random;

import org.colinvella.fancyfish.creativetab.ModCreativeTabs;
import org.colinvella.fancyfish.entity.ModEntityRegistry;
import org.colinvella.fancyfish.entity.fish.FishEntity;
import org.colinvella.fancyfish.entity.fish.MoorishIdolEntity;

import net.minecraft.block.Block;
import net.minecraft.block.material.Material;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.properties.PropertyBool;
import net.minecraft.block.properties.PropertyInteger;
import net.minecraft.block.state.BlockState;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.init.Blocks;
import net.minecraft.item.ItemStack;
import net.minecraft.util.AxisAlignedBB;
import net.minecraft.util.BlockPos;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.EnumWorldBlockLayer;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

public class KelpBlock extends ModBlock
{
	public static final String ID = "Kelp";
	
	// properties mapped to meta data
    public static final PropertyInteger PROPERTY_LEVEL = PropertyInteger.create("level", 0, 15);
	public static final PropertyBool PROPERTY_TIP = PropertyBool.create("tip");
	  
	public KelpBlock()
	{
		super(ID, Material.water, ModCreativeTabs.FancyFishTab);
		this.setLightOpacity(0); // light passes through
		this.setTickRandomly(true);
	}

	// override: return false as it is an irregular block
	@Override
	public boolean isOpaqueCube()
	{
		return false;
	}
	
	// override: treat block as translucent due to light shade
	@Override
	public EnumWorldBlockLayer getBlockLayer()
	{
		return EnumWorldBlockLayer.TRANSLUCENT;
	}
	
	// entities can pass through block
	@Override
    public AxisAlignedBB getCollisionBoundingBox(World worldIn, BlockPos pos, IBlockState state)
    {
        return null;
    }
	
    // Whether this Block is solid on the given Side
    @Override
    public boolean isBlockSolid(IBlockAccess worldIn, BlockPos pos, EnumFacing side)
    {
        Material material = worldIn.getBlockState(pos).getBlock().getMaterial();
        
		if (material == this.blockMaterial || material == Material.water)
        	return false;

        if (side == EnumFacing.UP)
        	return true;
        
        if (material == Material.ice)
        	return false;
        
        return super.isBlockSolid(worldIn, pos, side);
    }
	
	@Override
    @SideOnly(Side.CLIENT)
    public boolean shouldSideBeRendered(IBlockAccess worldIn, BlockPos pos, EnumFacing side)
    {
		Material material = worldIn.getBlockState(pos).getBlock().getMaterial();
		
		if (material == this.blockMaterial || material == Material.water)
			return false;
		
        if (side == EnumFacing.UP)
        	return true;

        return super.shouldSideBeRendered(worldIn, pos, side);
    }
	
	@Override
	public boolean canPlaceBlockAt(World world, BlockPos pos)
	{
		// target block must be water
		Block block = world.getBlockState(pos).getBlock();
		if (block != Blocks.water)
			return false;

		// block on which kelp grows must be dirt, sand or clay
		if (pos.getY() < 1)
			return false;

		Block blockBelow = world.getBlockState(pos.offset(EnumFacing.DOWN)).getBlock();
		if (!canPlaceKelpOn(blockBelow))
			return false;

		// ensure there is no air or waterlilly above to limit kelp to one block away from water surface
		if (pos.getY() >= world.getHeight() - 1)
			return false;

		Block blockAbove = world.getBlockState(pos.offset(EnumFacing.UP)).getBlock();
		return blockAbove != Blocks.air && blockAbove != Blocks.waterlily;
	};

	@Override
	public void onBlockAdded(World world, BlockPos pos, IBlockState blockState)
	{
		if (world.isRemote)
			return;
		
		IBlockState kelpState = blockState.withProperty(PROPERTY_TIP, false);
		IBlockState kelpTipState = blockState.withProperty(PROPERTY_TIP, true);

		// make kelp the tip variant, unless there is more kelp above it
		if (pos.getY() < world.getHeight() - 1)
		{
			BlockPos posAbove = pos.offset(EnumFacing.UP);
			Block blockAbove = world.getBlockState(posAbove).getBlock();
			if (blockAbove != ModBlockRegistry.Kelp)
			{
				world.setBlockState(pos, kelpTipState);
			}
		}
	};
	
	// break kelp if not surrounded by still water or supporting solid block
	@Override public void onNeighborBlockChange(World world, BlockPos pos, IBlockState blockState, Block neighborBlock)
	{
		if (world.isRemote)
			return;

		IBlockState kelpState = blockState.withProperty(PROPERTY_TIP, false);
		IBlockState kelpTipState = blockState.withProperty(PROPERTY_TIP, true);

		BlockPos posAbove = pos.offset(EnumFacing.UP);
		Block blockAbove = world.getBlockState(posAbove).getBlock();

		// check block above and change this kelp block to the tipped variant if needed
		if (blockAbove == ModBlockRegistry.Kelp && blockState != kelpState)
		{
			world.setBlockState(pos, kelpState);
		}
		// if kelp exposed to surface or waterlilly, break it
		else if (blockAbove == Blocks.air || blockAbove == Blocks.waterlily)
		{
			breakKelp(world, pos, true);
		}
		// check block above and change this kelp block to the non-tipped variant if needed
		else if (blockAbove != ModBlockRegistry.Kelp && blockState != kelpTipState)
		{
			world.setBlockState(pos, kelpTipState);
		}

		// if block below is no longer kelp or good ground, then break this kelp block
		BlockPos posBelow = pos.offset(EnumFacing.DOWN);
		Block blockBelow = world.getBlockState(posBelow).getBlock();
		if (!canPlaceKelpOn(blockBelow))
		{
			breakKelp(world, pos, true);
		}
	}
	
	@Override
	public void breakBlock(World world, BlockPos pos, IBlockState blockState)
	{
		breakKelp(world, pos, false);
		super.breakBlock(world, pos, blockState);
	};

	// passing through kelp slows entities down
	@Override
    public void onEntityCollidedWithBlock(World world, BlockPos pos, IBlockState state, Entity entity)
    {
        entity.motionX *= 0.6D;
        entity.motionY *= 0.6D;
        entity.motionZ *= 0.6D;
    }
	
	@Override
	public void randomDisplayTick(World world, BlockPos pos, IBlockState state, Random random)
	{
		// this runs only on client
		// find kelp tip
		pos = this.getPosAboveKelp(world, pos);

		Block blockAboveTip = world.getBlockState(pos).getBlock();
		// nowhere to emit bubbles
		if (blockAboveTip != Blocks.water)
			return;
		
		// 10% chance of spawning bubble
		if (random.nextInt(10) > 0)
			return;

		double bubbleX = pos.getX() + random.nextDouble();
		double bubbleY = pos.getY();
		double bubbleZ = pos.getZ() + random.nextDouble();
		world.spawnParticle(EnumParticleTypes.WATER_BUBBLE, bubbleX, bubbleY, bubbleZ, 0, 1, 0);
	};
	
	@Override 
	public void updateTick(World world, BlockPos pos, IBlockState state, Random random)
	{
		// note: this runs only on server
		
		// 25% chance of growing
		if (random.nextInt(4) > 0)
			return;
		
		// find kelp tip
		pos = this.getPosAboveKelp(world, pos);

		Block blockAboveTip = world.getBlockState(pos).getBlock();
		// nowhere to grow or emit bubbles
		if (blockAboveTip != Blocks.water)
			return;

		// cannot grow to surface or if there is waterlilly
		Block blockTwoAboveTip = world.getBlockState(pos.offset(EnumFacing.UP)).getBlock();
		if (blockTwoAboveTip == Blocks.air || blockTwoAboveTip == Blocks.waterlily)
			return;

		// grow kelp by a block
		world.setBlockState(pos, ModBlockRegistry.Kelp.getDefaultState().withProperty(PROPERTY_TIP, true));
	};
		
	// getStateFromMeta, getMetaFromState are used to inter-convert between the block's property values and
	//   the stored metadata (which must be an integer in the range 0 - 15 inclusive)
	@Override
	public IBlockState getStateFromMeta(int meta)
	{
		boolean isTip = meta > 0;
		return this.getDefaultState().withProperty(PROPERTY_TIP, isTip);
	}
	
	@Override
	public int getMetaFromState(IBlockState state)
	{
		boolean isTip = (Boolean)state.getValue(PROPERTY_TIP);	
		return isTip ? 1: 0;
	}

	// necessary to define which properties your blocks use
	// will also affect the variants listed in the blocskstates model file
	@Override
	protected BlockState createBlockState()
	{
	    return new BlockState(this, new IProperty[] {PROPERTY_LEVEL, PROPERTY_TIP});
	}
	
	private BlockPos getPosAboveKelp(World world, BlockPos pos)
	{
		while (world.getBlockState(pos).getBlock() == ModBlockRegistry.Kelp)
			pos = pos.offset(EnumFacing.UP);
		return pos;
	}
	
	private boolean canPlaceKelpOn(Block block)
	{
		return block == Blocks.sand || block == Blocks.dirt || block == Blocks.clay || block == ModBlockRegistry.Kelp;
	}
	
	private void breakKelp(World world, BlockPos pos, boolean spawnKelpItem)
	{
		Block block = world.getBlockState(pos).getBlock();
		
		// replace kelp with still water
		world.setBlockState(pos, Blocks.water.getDefaultState().withProperty(PROPERTY_LEVEL, 15));
		
		double midX = pos.getX() + 0.5;
		double midY = pos.getY() + 0.5;
		double midZ = pos.getZ() + 0.5;
		
		// spawn kelp as an item if requested
		if (spawnKelpItem)
		{
			world.spawnEntityInWorld(
				new EntityItem(world, midX, midY, midZ, new ItemStack(ModBlockRegistry.Kelp)));
		}
		
		// 20% chance spawn a random fish (hiding in kelp)
		if (this.RANDOM.nextInt(5) > 0)
			return;
		
		// get list of fish entity classes
		List<Class<? extends FishEntity>> fishEntityClasses
			= ModEntityRegistry.getModFishEntityClasses();
		
		// choose one at random
		Class<? extends FishEntity> fishEntityClass
			= fishEntityClasses.get(this.RANDOM.nextInt(fishEntityClasses.size()));

		// spawn fish
		try {
			FishEntity fishEntity = fishEntityClass.getConstructor(new Class[] {World.class}).newInstance(world);
			fishEntity.setPosition(midX, midY, midZ);
			world.spawnEntityInWorld(fishEntity);
		} catch (Exception exception)
		{
			logger.error("Cannot spawn fish. Reason: " + exception.getMessage());
		}
	}
}
