package org.colinvella.fancyfish.tileentity;

import org.colinvella.fancyfish.block.FishTankBlock;
import org.colinvella.fancyfish.entity.fish.FishEntity;
import org.colinvella.fancyfish.item.CapturedFishItem;
import org.colinvella.fancyfish.item.LiveFishItem;
import org.colinvella.fancyfish.item.ModItemRegistry;

import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.Packet;
import net.minecraft.network.play.server.S35PacketUpdateTileEntity;
import net.minecraft.world.World;

public class FishTankTileEntity extends ModInventoryTileEntity
{
	public FishTankTileEntity()
	{
		super();
		
		this.capturedFishStates = new CapturedFishState[9];
		for (int slotIndex = 0; slotIndex < 9; slotIndex++)
		{
			this.capturedFishStates[slotIndex] = new CapturedFishState();
		}
	}
	
	public CapturedFishState[] getCapturedFishStates()
	{
		return this.capturedFishStates;
	}
	
	@Override
	public int getSizeInventory()
	{
		return 9;
	}
	
	@Override
	public int getInventoryStackLimit()
	{
		return 1;
	}
	
	// stop hoppers from putting anything other than fish
	@Override
	public boolean isItemValidForSlot(int index, ItemStack stack)
	{
		if (stack == null)
			return false;
		
		return stack.getItem() instanceof CapturedFishItem;
	}
	
	@Override
	public void update()
	{
		if (this.worldObj.isRemote)
		{
			// client side
			
			// move fish around
			double deltaTime = 0.05; // assume 20 ticks/s
			for (int slotIndex = 0; slotIndex < 9; slotIndex++)
			{
				this.capturedFishStates[slotIndex].update(deltaTime);
			}
		}
		else
		{
			// server side
			replaceCapturedFishWithLiveFish();
			
			IBlockState currentState = this.worldObj.getBlockState(pos);
			IBlockState filledState = currentState.withProperty(FishTankBlock.PROPERTY_FILLED, true);
			if (currentState != filledState && this.hasFish())
				ejectFishEntities();

		}
		
		super.update();
	}
	
	// sync tile entities between client and server (for TESR animation)
	
	// serialise
    @Override
    public Packet getDescriptionPacket()
    {
        NBTTagCompound syncData = new NBTTagCompound();
        this.writeToNBT(syncData);
        return new S35PacketUpdateTileEntity(this.pos, this.getBlockMetadata(), syncData);
    }
    
    // deserialise
    @Override
    public void onDataPacket(NetworkManager net, S35PacketUpdateTileEntity pkt)
    {
	    this.readFromNBT(pkt.getNbtCompound());
    }		 
	
	public void ejectFishEntities()
	{
		// scan fish tank for captured fish items, remove them and spawn corresponding fish entities on top of tank block
		int slotCount = this.getSizeInventory();
		for (int slotIndex = 0; slotIndex < slotCount; slotIndex++)
		{
			ItemStack itemStack = this.getStackInSlot(slotIndex);
			if (itemStack == null)
				continue;		
			
			// replace captured fish with live fish and drop fish bowl
			Item item = itemStack.getItem();			

			// if tank has no water, clear slot and spawn corresponding fish entity
			ItemStack liveFishItemStack = this.getStackInSlot(slotIndex);
			LiveFishItem liveFishItem = (LiveFishItem)liveFishItemStack.getItem();
			Class<? extends FishEntity> fishEntityClass = liveFishItem.getFishEntityClass();
			
			FishEntity fishEntity;
			try
			{
				fishEntity = fishEntityClass.getConstructor(World.class).newInstance(this.worldObj);
			}
			catch (Exception exception)
			{
				// clear slot
				this.setInventorySlotContents(slotIndex, null);
				
				continue;
			}
		
			// retrieve tag in case fish was originally assigned one
			NBTTagCompound tagCompound = liveFishItemStack.getTagCompound();
			String fishEntityTag = tagCompound == null ? null : tagCompound.getString("tag");
			
			// restore custom name tag if applicable
			if (fishEntityTag != null)
				fishEntity.setCustomNameTag(fishEntityTag);

			// set position from hit coordinates and spawn fish
			fishEntity.setPosition(pos.getX() + 0.5, pos.getY() + 1.5, pos.getZ() + 0.5);
			this.worldObj.spawnEntityInWorld(fishEntity);
			
			// clear slot
			this.setInventorySlotContents(slotIndex, null);
		}		

		// trigger server -> client synch
		this.worldObj.markBlockForUpdate(pos);
		this.markDirty();
	}
	
	private boolean hasFish()
	{
		int slotCount = this.getSizeInventory();
		for (int slotIndex = 0; slotIndex < slotCount; slotIndex++)
			if (this.getStackInSlot(slotIndex) != null)
				return true;
		return false;
	}
	
	private void replaceCapturedFishWithLiveFish()
	{
		// scan fish tank for captured fish items, replace them with live fish equivalents
		// and drop corresponding empty fish bowls for player to pick up (automatic by proximity)
		int slotCount = this.getSizeInventory();
		for (int slotIndex = 0; slotIndex < slotCount; slotIndex++)
		{
			ItemStack itemStack = this.getStackInSlot(slotIndex);
			if (itemStack == null)
				continue;		
			
			// replace captured fish with live fish and drop fish bowl
			Item item = itemStack.getItem();			
			if (item instanceof CapturedFishItem)
			{
				CapturedFishItem capturedFishItem = (CapturedFishItem)item;
				LiveFishItem liveFishItem = ModItemRegistry.getLiveFromCaptured(capturedFishItem);
				if (liveFishItem != null)
				{
					// retrieve tag from captured fish in case fish was originally assigned one
					NBTTagCompound tagCompound = itemStack.getTagCompound();

					ItemStack liveFishItemStack = new ItemStack(liveFishItem);
					// restore custom name tag if applicable
					liveFishItemStack.setTagCompound(tagCompound);
					
					this.setInventorySlotContents(slotIndex, liveFishItemStack);
					this.markDirty();
				}
				
				// drop empty fish bowl
				this.worldObj.spawnEntityInWorld(new EntityItem(worldObj,
						pos.getX() + 0.5, pos.getY() + 0.5, pos.getZ() + 0.5,
					new ItemStack(ModItemRegistry.FishBowl)));
			}
		}		
	}
	
	private CapturedFishState[] capturedFishStates;
}
